// code based on kdbus/test/test-kdbus-benchmark.c, Copyright belongs to original authors
/*
 * kdbus is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 */

// additional code:
// Copyright (C) 2014 Stefan Westerfeld <stefan@space.twc.de>


#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stddef.h>
#include <stdbool.h>
#include <assert.h>
#include <errno.h>
#include <poll.h>
#include <math.h>
#include <pthread.h>

#include "kdbus.h"
#include "kdbus-util.h"

#define KBUILD_MODNAME  "kdbus"
#define MSG_TIMEOUT_NS  (25 * 1000ULL * 1000ULL * 1000ULL)
#define KDBUS_POOL_SIZE (16*1024*1024)
#define MAX_SIZE      (4 * 1024 * 1024)

#define SERVICE_NAME    "foo.kdbench.echo"

static void
die (const char *why)
{
  fputs (why, stderr);
  exit (1);
}

static void
set_cpu_affinity (int cpu)
{
  pthread_t thread = pthread_self();
  cpu_set_t cpuset;
  if (cpu >= 0 && cpu < CPU_SETSIZE)
    {
      CPU_ZERO (&cpuset);
      CPU_SET (cpu, &cpuset);
      if (pthread_setaffinity_np (thread, sizeof (cpu_set_t), &cpuset) == 0)
        return; /* success */
    }
  die ("set_cpu_affinity failed\n");
}

static void
poll_wait (int fd)
{
  struct pollfd poll_fds[1];
  poll_fds[0].fd = fd;
  poll_fds[0].events = POLLIN;
  poll_fds[0].revents = 0;
  for (;;)
    {
      if (poll (poll_fds, 1, -1) > 0)
        {
          return;
        }
    }
}

static uint64_t
get_id_xxx (struct conn *my_conn)
{
  if (my_conn->id == 1)
    return 2;
  if (my_conn->id == 2)
    return 1;
  die ("expected connection id 1 or 2\n");
}

static char quit_now_payload[1] = { 0 };

static void
send_request (struct conn *conn, uint64_t dst_id, void *kdbus_buffer, int quit_now, size_t content_size, unsigned char *content, bool use_memfd, int memfd)
{
  struct kdbus_msg   *msg;
  struct kdbus_item  *item;
  uint64_t            size;
  int                 ret;

  size = sizeof (struct kdbus_msg);

  if (use_memfd)
    size += KDBUS_ITEM_SIZE(sizeof (struct kdbus_memfd));
  else
    size += KDBUS_ITEM_SIZE (sizeof (struct kdbus_vec));

  msg = alloca (size);
  if (!msg)
    die ("unable to alloca()!?\n");

  memset (msg, 0, size);
  msg->size = size;
  msg->flags = KDBUS_MSG_FLAGS_EXPECT_REPLY|KDBUS_MSG_FLAGS_SYNC_REPLY; // synchronous call, block for result
  msg->src_id = conn->id;
  msg->dst_id = dst_id;
  msg->payload_type = KDBUS_PAYLOAD_DBUS;
  msg->timeout_ns = MSG_TIMEOUT_NS;
  msg->cookie = 1;

  item = msg->items;

  if (use_memfd)
    {
      if (quit_now)
        die ("send_request: quit_now should never use memfd payload.");

      ret = ioctl (memfd, KDBUS_CMD_MEMFD_SEAL_SET, false);
      if (ret < 0)
        die ("memfd unsealing failed\n");

      uint64_t sz = content_size;
      ret = ioctl (memfd, KDBUS_CMD_MEMFD_SIZE_SET, &sz);
      if (ret < 0)
        die ("memfd resize failed\n");

      unsigned char *buf = mmap (NULL, content_size, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, 0);

      if (buf == MAP_FAILED)
        die ("mmap() failed in client process");

      memset (buf, 0x80, content_size); // generate some data to send
      munmap (buf, content_size);

      ret = ioctl (memfd, KDBUS_CMD_MEMFD_SEAL_SET, true);
      if (ret < 0)
        die ("memfd sealing failed\n");

      item->type = KDBUS_ITEM_PAYLOAD_MEMFD;
      item->size = KDBUS_ITEM_HEADER_SIZE + sizeof (struct kdbus_memfd);
      item->memfd.size = content_size;
      item->memfd.fd = memfd;
    }
  else
    {
      item->type = KDBUS_ITEM_PAYLOAD_VEC;
      item->size = KDBUS_ITEM_HEADER_SIZE + sizeof(struct kdbus_vec);

      if (quit_now)
        {
          item->vec.address = (uintptr_t) quit_now_payload;
          item->vec.size    = sizeof (quit_now_payload);
        }
      else
        {
          item->vec.address = (uintptr_t) content;
          item->vec.size    = content_size;

          memset (content, 0x80, content_size); // generate some data to send
        }
    }


  ret = ioctl (conn->fd, KDBUS_CMD_MSG_SEND, msg);
  if (ret)
    {
      fprintf(stderr, "error sending message: %d err %d (%m)\n", ret, errno);
      exit (1);
    }

#if 0
  printf ("got reply, offset is %llu\n", msg->offset_reply);
#endif
  struct kdbus_msg   *reply_msg = msg->offset_reply + kdbus_buffer;

  item = reply_msg->items;
  assert (item->type == KDBUS_ITEM_PAYLOAD_OFF);
#if 0
  printf ("size=%lld\n", item->vec.size);
  for (int i = 0; i < item->vec.size; i++)
    printf ("reply|%d|%c\n", i, ((char *) reply_msg) [item->vec.address + i]);
#endif

  uint64_t off = msg->offset_reply;
  ret = ioctl (conn->fd, KDBUS_CMD_FREE, &off);
  if (ret < 0)
    {
      fprintf(stderr, "error free message: %d (%m)\n", ret);
      exit (1);
    }
}

static char reply_payload[] = { 'r', 'e', 'p', 'l', 'y' };

static void
send_reply (struct conn *conn, uint64_t dst_id)
{
  struct kdbus_msg   *msg;
  struct kdbus_item  *item;
  uint64_t            size;
  int                 ret;

  size = sizeof (struct kdbus_msg);
  size += KDBUS_ITEM_SIZE (sizeof (struct kdbus_vec));

  msg = alloca (size);
  if (!msg)
    die ("unable to alloca()!?\n");

  memset (msg, 0, size);
  msg->size = size;
  msg->flags = 0;
  msg->src_id = conn->id;
  msg->dst_id = dst_id;
  msg->cookie = 1;
  msg->payload_type = KDBUS_PAYLOAD_DBUS;
  msg->cookie_reply = 1;

  item = msg->items;

  item->type = KDBUS_ITEM_PAYLOAD_VEC;
  item->size = KDBUS_ITEM_HEADER_SIZE + sizeof(struct kdbus_vec);
  item->vec.address = (uintptr_t) reply_payload;
  item->vec.size = sizeof (reply_payload);

  ret = ioctl (conn->fd, KDBUS_CMD_MSG_SEND, msg);
  if (ret)
    {
      fprintf(stderr, "error sending reply: %d err %d (%m)\n", ret, errno);
      exit (1);
    }
}

static double
gettime()
{
  struct timeval tv;
  gettimeofday (&tv, 0);

  return ((double) tv.tv_sec) + (((double) tv.tv_usec) / 1000000.0);
}

static double
client_loop (struct conn *client_conn, uint64_t server_id, void *kdbus_buffer, size_t content_size, unsigned char *content, int n_calls, bool use_memfd, int memfd)
{
  int i;

  double start = gettime();
  for (i = 0; i < n_calls; i++)
    {
      send_request (client_conn, server_id, kdbus_buffer, /* quit_now */ 0, content_size, content, use_memfd, memfd);
    }
  double end = gettime();

  return end - start;
}

static void
run_client (const char *bus)
{
  sleep (1);  // wait for server process to connect to bus

  struct conn *client_conn = kdbus_hello (bus, 0);
  if (!client_conn)
    die ("failed to connect bus in client\n");

  //!! add_match_empty (client_conn->fd);

  void *kdbus_buffer = mmap (NULL, KDBUS_POOL_SIZE, PROT_READ, MAP_SHARED, client_conn->fd, 0);
  assert (kdbus_buffer);

  // we can reuse the same memfd for all requests
  struct kdbus_cmd_memfd_make mfd = {};

  mfd.size = sizeof (struct kdbus_cmd_memfd_make);
  mfd.file_size = 0;

  int ret = ioctl (client_conn->fd, KDBUS_CMD_MEMFD_NEW, &mfd);
  if (ret < 0)
    die ("KDBUS_CMD_MEMFD_NEW failed.\n");

  int memfd = mfd.fd;

  uint64_t server_id = get_id_xxx (client_conn);

  long n_cpus = sysconf (_SC_NPROCESSORS_ONLN);
  printf ("%ld cpu(s) are available on this system.\n\n", n_cpus);

  for (int cpu = 0; cpu < (n_cpus > 1 ? 2 : 1); cpu++)
    {
      if (cpu == 0)
        printf ("*** single cpu performance:\n\n");
      else
        printf ("*** dual cpu performance:\n\n");

      printf ("SIZE\tCOPY\tMEMFD\n\n");

      set_cpu_affinity (cpu);

      size_t content_size;
      for (content_size = 1; content_size <= MAX_SIZE; content_size *= 2)
        {
          unsigned char *content = malloc (content_size);
          if (!content)
            die ("content malloc() failed.\n");

          int calls_per_sec[2];

          for (int use_memfd = 0; use_memfd <= 1; use_memfd++)
            {
              int n_calls = 10;
              double time = 0;

              for (;;)
                {
                  time = client_loop (client_conn, server_id, kdbus_buffer, content_size, content, n_calls, use_memfd, memfd);
                  if (time > 0.1)
                    break;

                  n_calls *= 2;
                }
              calls_per_sec[use_memfd] = lrint (n_calls / time);
            }
          printf ("%zd\t%d\t%d\n", content_size, calls_per_sec[0], calls_per_sec[1]);

          free (content);
        }
      printf ("\n");
    }

  send_request (client_conn, server_id, kdbus_buffer, /* quit_now */ 1, 0, NULL, false, memfd);

  assert (memfd >= 0);
  close (memfd);
}

static void
validate_content (size_t content_size, unsigned char *content)
{
#if 0
  assert (content_size > 0);
  for (size_t i = 0; i < content_size; i++)
    {
      if (content[i] != 0x80)
        die ("validate_data: validation error");
      // printf ("data|%zd|%x\n", i, content[i]);
    }
#endif
}

static void
run_server (const char *bus)
{
  struct conn *server_conn = kdbus_hello (bus, 0);
  if (!server_conn)
    die ("failed to connect bus in server\n");

  //!! upload_policy(server_conn->fd, SERVICE_NAME);
  //!! add_match_empty(server_conn->fd);
  //!! name_acquire(server_conn, SERVICE_NAME, 0);

  uint64_t client_id = get_id_xxx (server_conn);

  set_cpu_affinity (0);
  for (;;)
    {
      poll_wait (server_conn->fd);      // wait for client thread requests

      struct kdbus_cmd_recv recv = {};
      int ret = ioctl (server_conn->fd, KDBUS_CMD_MSG_RECV, &recv);
      if (ret < 0)
        {
          fprintf(stderr, "error receiving message: %d (%m)\n", ret);
          exit (1);
        }

      struct kdbus_msg *msg;
      msg = (struct kdbus_msg *)(server_conn->buf + recv.offset);

      struct kdbus_item *item;
      item = msg->items;

      assert (item->type == KDBUS_ITEM_PAYLOAD_OFF || item->type == KDBUS_ITEM_PAYLOAD_MEMFD);
#if 0
      printf ("size=%lld\n", item->vec.size);
      for (int i = 0; i < item->vec.size; i++)
        printf ("data|%d|%c\n", i, ((char *) msg) [item->vec.address + i]);
#endif
      int quit_now = (item->type == KDBUS_ITEM_PAYLOAD_OFF && item->vec.size == 1 && ((char *) msg) [item->vec.address] == 0);
      if (item->type == KDBUS_ITEM_PAYLOAD_OFF && !quit_now)
        {
          validate_content (item->vec.size, (unsigned char *) msg + item->vec.address);
        }
      else if (item->type == KDBUS_ITEM_PAYLOAD_MEMFD)
        {
          unsigned char *buf = mmap (NULL, item->memfd.size, PROT_READ, MAP_SHARED, item->memfd.fd, 0);

          if (buf == MAP_FAILED)
            die ("mmap() failed in server process");

          validate_content (item->memfd.size, buf);
          munmap (buf, item->memfd.size);

          close (item->memfd.fd);
        }

      ret = ioctl (server_conn->fd, KDBUS_CMD_FREE, &recv.offset);
      if (ret < 0)
        {
          fprintf(stderr, "error free message: %d (%m)\n", ret);
          exit (1);
        }

      send_reply (server_conn, client_id);

      if (quit_now)
        break;
    }
}

int
main()
{
  struct {
    struct kdbus_cmd_make head;

    /* bloom size item */
    struct {
      uint64_t size;
      uint64_t type;
      struct kdbus_bloom_parameter bloom;
    } bs;

    /* name item */
    uint64_t n_size;
    uint64_t n_type;
    char name[64];
  } bus_make;

  printf("-- kdbench version %s\n", KDBENCH_VERSION);
  printf("-- opening /dev/" KBUILD_MODNAME "/control\n");
  int fdc = open("/dev/" KBUILD_MODNAME "/control", O_RDWR|O_CLOEXEC);
  if (fdc < 0)
    {
      fprintf(stderr, "--- error %d (%m)\n", fdc);
      return EXIT_FAILURE;
    }

  memset(&bus_make, 0, sizeof(bus_make));
  bus_make.bs.size = sizeof(bus_make.bs);
  bus_make.bs.type = KDBUS_ITEM_BLOOM_PARAMETER;
  bus_make.bs.bloom.size = 64;
  bus_make.bs.bloom.n_hash = 1;

  snprintf(bus_make.name, sizeof(bus_make.name), "%u-kbenchbus", getuid());
  bus_make.n_type = KDBUS_ITEM_MAKE_NAME;
  bus_make.n_size = KDBUS_ITEM_HEADER_SIZE + strlen(bus_make.name) + 1;

  bus_make.head.size = sizeof(struct kdbus_cmd_make) + sizeof(bus_make.bs) + bus_make.n_size;

  char *bus;
  if (asprintf(&bus, "/dev/" KBUILD_MODNAME "/%s/bus", bus_make.name) < 0)
    return EXIT_FAILURE;

  int ret;

  printf("-- creating bus '%s'\n", bus_make.name);
  ret = ioctl(fdc, KDBUS_CMD_BUS_MAKE, &bus_make);
  if (ret)
    {
      fprintf(stderr, "--- error %d (%m)\n", ret);
      return EXIT_FAILURE;
    }

  pid_t pid = fork();
  if (pid == 0)
  {
    run_server (bus);

    _exit (0);
  }

  run_client (bus);

  if (waitpid (pid, NULL, 0) != pid)
    die ("waitpid() failed.\n");

  return 0;

}
